// FILENAME: HTree.C
// ------------------------------------------------------------
// Copyright 1998, The University of California at Irvine
// ------------------------------------------------------------
// Written by Kaushik Chakrabarti
// Implementation of the the hybrid tree.
// implements functions to search the hybrid tree (range
// and nn-queries, insert a point into the hybrid tree, 
// delete a point from the tree, build the tree from a data
// file provided it is the following format: <id> <float1> ... <floatk>
// (corr. to the k dims), dump an existing htree into a file
// and load the htree back to memory from the dumpfile
// ------------------------------------------------------------



#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <assert.h>
#include "HTree.h"



int HTree::Scan(Query *query)
{
  int i;
  int how_many, rest;
  PQ *src_queue, *dest_queue, *startQueue;
  Result *src_result;

  if (query->type == Query::RANGEQUERY)  // if it is a range query
    {
      assert(!(query->feedback)); // no question of feedback
      assert(!(query->getmore));  // no question of getmore
      assert(query->startQuery);  // only single point queries
      if (query->use_function) // use user specified function
	return Search(&(query->startQuery), query->range, query->dist_func, query->diskacc, query->startResult, query->max_returned);
      else        // use the dist metric specified by the user L1, L2 or LMAX
	return Search(&(query->startQuery), query->range, query->dist_metric, query->diskacc, query->startResult, query->max_returned);
    }
  // END OF RANGE QUERIES
  else    // NN queries
    {
      assert(query->type == Query::NNQUERY); // for now we are handling only NNqueries
      
      // **************** STARTING QUERY ******************

      if(!(query->feedback)) // it is not feedback, it is a fresh query
	{
          query->iteration=0;   // intitialized iteration
	  startQueue=query->startQueue;
	  if (!(query->getmore))
	    {
	      assert(query->startQueue.empty());  // startQueue should be empty
	      assert(query->startResult->numObjects==0); // no results
              // still, make sure it is empty
              query->startResult->numObjects=0;
	      // push in the root
	      startQueue->push(PQEntry(HTreeRoot, NODE, rootRect, 0.0));
	    }
	  else // getting more stuff, continuing from previous query
	    {
	      // make sure startQueue is not empty
	      assert(!(query->startQueue.empty()));
	    }
	  if (query->use_function) // use the user_specified distance function
	    return (GetkNext(startQueue, &(query->startQuery), query->dist_func, query->diskacc, query->k, query->startResult)); // k nearest neighbors 
	  else  // use the distance metric specified by the user
	    return (GetkNext(startQueue, &(query->startQuery), query->dist_metric, query->diskacc, query->k, query->startResult)); // k nearest neighbors
	}

      // **************** REFINED QUERY (FEEDBACK) ******************

      else // feedback
	{
	  if (!(query->getmore)) (query->iteration)++;   // incr. the number of iterations of feedback

	  // select the src_queue and dest_queue
	  // swapping them between startQueue and lastQueue from iteration to iteration
	  if (((query->iteration)%2)!=0)
	    {
	      src_queue=query->startQueue;
	      dest_queue=query->lastQueue;
	    }
	  else
	    {
	      src_queue=query->lastQueue;
	      dest_queue=query->startQueue;
	    }
	  assert((dest_queue)->empty());  // at this point, dest queue must be empty

          if (query->iteration==1)  // this is the first iteration of feedback
            src_result=query->startResult;
          else
            src_result=query->lastResult;
	  
          // push in all the objects in src_result into the dest queue
          // first,recomputing distance based on new multpoint distance
          // function
          if (query->use_function) // use the user-defined function
	    {
	      for(i=0; i<src_result->numObjects; i++)
		dest_queue->push(PQEntry((Node *)(src_result->objectList[i].id), OBJECT, (src_result->objectList[i].point).ToRect(), (query->lastQuery).Distance(&(src_result->objectList[i].point), query->dist_func)));
	    }
          else // use the dist-metric
	    {
	      for(i=0; i<src_result->numObjects; i++)
		dest_queue->push(PQEntry((Node *)(src_result->objectList[i].id), OBJECT, (src_result->objectList[i].point).ToRect(), (query->lastQuery).Distance(&(src_result->objectList[i].point), query->dist_metric)));
	    }
	  
          src_result->numObjects=0;   // all objects from src_result has been transferred
	  
	  // lastResult is ready to be filled up with new results
	  assert(query->lastResult->numObjects==0);
          if (!(query->getmore))
	    {
	      if (query->use_function) // use the user-defined function
		  PushAll(src_queue, dest_queue, &(query->lastQuery), query->dist_func);
	      else PushAll(src_queue, dest_queue, &(query->lastQuery), query->dist_metric);
	    }
	      
	  if (query->use_function) return (GetkNext(dest_queue, &(query->lastQuery), query->dist_func, query->diskacc, query->k, query->lastResult));
          return (GetkNext(dest_queue, &(query->lastQuery), query->dist_metric, query->diskacc, query->k, query->lastResult));
	}
    }
}


// ------------------- NN SEARCH FOR START QUERIES ---------------------
// ------------------- NN SEARCH FOR START QUERIES ---------------------
// ------------------- NN SEARCH FOR START QUERIES ---------------------
// NN search for first time searches
// feedback code is above
// for multipoint queries
// Search (plain partial reuse) used this function
// NN query using user specified dist func
// NEAREST NEIGHBOR SEARCH, returns the oid of the NN object

int HTree::GetkNext(PQ *queue, MultiPointQuery *MPQ, float (*User_Distance) (Point *, Point *), long *diskacc, int k, Result *res)
{
	register int i;
	Rect childRectlist[OFNODECARD];
	Rect MOchildRectlist[MONODECARD];
	int sofar=0;

	assert(queue);
	assert(MPQ);


	while(!(queue->empty()))
	  {
	    PQEntry top=queue->top();
	    queue->pop();

	    if (top.type == NODE)
	      {
		Node *node=top.n;
		// if (IsPageFault(node)) diskacc[node->level+1]++; // to include buffering effect, uncomment this line and comment the next line
		diskacc[node->level+1]++;
		if (node->level > 0) // this is an internal node in the tree 
		  {
		    if (node->type== Node::MONODE)   // this is an ordinary node
		      {
			MONode *N=(MONode *)node;
			N->FillChildList(0, MPQ, &(top.BR), queue, User_Distance);
		      }
		    else // this is an OFnode (type is OVERLAP_FREE)
		      {
			assert(node->type== Node::OFNODE);
			OFNode *N=(OFNode *)node;
			N->FillChildList(0, MPQ, &(top.BR), queue, User_Distance);
		      }
		  }
		else // this is a leaf node 
		  {
		    DataNode *N=(DataNode *)node;
		    for (i=0; i<N->count; i++)
		      queue->push(PQEntry((N->branch[i]).child, OBJECT, (N->branch[i].point).ToRect(), MPQ->Distance(&(N->branch[i].point), User_Distance)));
		  }
	      }
	    else
	      {
		diskacc[0]++;
		assert(top.type == OBJECT);
		res->numObjects++;
		(res->objectList[sofar]).id=(long)top.n;
		(res->objectList[sofar]).point=(top.BR).ToPoint();
		(res->objectList[sofar]).distance=top.distance;
		sofar++;
		if (sofar ==k ) // got k nearest neigbors
		return sofar;
	      }
	  }
	return -1;
}


// Search (plain partial reuse) used this function
// NN query using user specified dist func
// NEAREST NEIGHBOR SEARCH, returns the oid of the NN object

int HTree::GetkNext(PQ *queue, MultiPointQuery *MPQ, int dist_func, long *diskacc, int k, Result *res)
{
	register int i;
	Rect childRectlist[OFNODECARD];
	Rect MOchildRectlist[MONODECARD];
	int sofar=0;

	assert(queue);
	assert(MPQ);


	while(!(queue->empty()))
	  {
	    PQEntry top=queue->top();
	    queue->pop();

	    if (top.type == NODE)
	      {
		Node *node=top.n;
		// if (IsPageFault(node)) diskacc[node->level+1]++; // to include buffering effect, uncomment this line and comment the next line
		diskacc[node->level+1]++;
		if (node->level > 0) // this is an internal node in the tree 
		  {
		    if (node->type== Node::MONODE)   // this is an ordinary node
		      {
			MONode *N=(MONode *)node;
			N->FillChildList(0, MPQ, &(top.BR), queue, dist_func);
		      }
		    else // this is an OFnode (type is OVERLAP_FREE)
		      {
			assert(node->type== Node::OFNODE);
			OFNode *N=(OFNode *)node;
			N->FillChildList(0, MPQ, &(top.BR), queue, dist_func);
		      }
		  }
		else // this is a leaf node 
		  {
		    DataNode *N=(DataNode *)node;
		    for (i=0; i<N->count; i++)
		      queue->push(PQEntry((N->branch[i]).child, OBJECT, (N->branch[i].point).ToRect(), MPQ->Distance(&(N->branch[i].point), dist_func)));
		  }
	      }
	    else
	      {
		diskacc[0]++;
		assert(top.type == OBJECT);
		res->numObjects++;
		(res->objectList[sofar]).id=(long)top.n;
		(res->objectList[sofar]).point=(top.BR).ToPoint();
		(res->objectList[sofar]).distance=top.distance;
		sofar++;
		if (sofar ==k ) // got k nearest neigbors
		return sofar;
	      }
	  }
	return -1;
}

// ----------------- FEEDBACK CODE BEGINS HERE -------------------
// ----------------- FEEDBACK CODE BEGINS HERE -------------------
// ----------------- FEEDBACK CODE BEGINS HERE -------------------
// ----------------- FEEDBACK CODE BEGINS HERE -------------------


int HTree::GetkNext(PQ *src_queue, PQ *dest_queue, MultiPointQuery *oldMPQ, MultiPointQuery *MPQ, float (*User_Distance) (Point *, Point *), long *diskacc, int k, Result *res)
{
  register int i, j, l, jj;
  Point p;
  int index_of_farthest_point;
  float dist, obj_dist, dist_from_orig_query, new_dist;
  float bound;
  Point object_point;
  Rect liveRegion;
  assert(src_queue);
  assert(dest_queue);
  assert(MPQ);
  int num_dist_comp=0;
  int max_num_dist_comp=src_queue->size();
  float smallest_dist, curr_radius;

  // assuming the sum of weights of any MPQ is 1.0

  float bound_diff=MPQ->Distance(oldMPQ, User_Distance);
  assert(!(src_queue->empty()));   // make sure that source queue is not empty
  assert(!(dest_queue->empty()));  // make sure that dest queue is not empty
  for(;;)
    {
      // bring stuff from the src_queue to the dest_queue
      while(!(src_queue->empty()))
	{
	  PQEntry src_top=src_queue->top();
	  PQEntry dest_top=dest_queue->top();
	  if (dest_top.distance <= (src_top.distance-bound_diff)) break;  // go to the explore phase
	  else  
	    {
	      // pop from src queue and push into dest queue, after computing the new distance
	      num_dist_comp++;
	      src_queue->pop();
	      if (src_top.type == OBJECT)
		{
		  object_point=(src_top.BR).ToPoint();
		  new_dist=MPQ->Distance(&object_point, User_Distance); // distance based on the new set of points
		  dest_queue->push(PQEntry(src_top.n, src_top.type, src_top.BR, new_dist));
		}
	      else
		{
		  if (src_top.els.elsvalid)
		    {
		      // use the distance from the live BB instead of the whole BB
		      liveRegion.DecodeLiveSpace(src_top.els.elsarray, &(src_top.BR));
		      // everything same as source queue, only distance is recomputed
		      new_dist=MPQ->Distance(&liveRegion, User_Distance);
		      dest_queue->push(PQEntry(src_top.n, src_top.type, src_top.BR, new_dist, src_top.count, src_top.els.elsarray));
		    }
		  else
		    {
		      new_dist=MPQ->Distance(&(src_top.BR), User_Distance);
		      dest_queue->push(PQEntry(src_top.n, src_top.type, src_top.BR, new_dist, src_top.count, NULL));
		    }
		}
	    }
	}

      // explore the first guy in the dest_queue 
      // if k objects already obtained, declare done
      // if not, go back to the bring phase
      if(!(dest_queue->empty()))
	{
	  PQEntry dest_top=dest_queue->top();
	  dest_queue->pop();
	  if (dest_top.type == NODE)
	    {
	      Node *node=dest_top.n;
	      diskacc[node->level+1]++;
	      if (node->level > 0) // this is an internal node in the tree 
		{
		  if (node->type== Node::MONODE)   // this is an ordinary node
		    {
		      MONode *N=(MONode *)node;
		      N->FillChildList(0, MPQ, &(dest_top.BR), dest_queue, User_Distance);
		    }
		  else // this is an OFnode (type is OVERLAP_FREE)
		    {
		      assert(node->type== Node::OFNODE);
		      OFNode *N=(OFNode *)node;
		      N->FillChildList(0, MPQ, &(dest_top.BR), dest_queue, User_Distance);
		    }
		}
	      else // this is a leaf node 
		{
		  DataNode *N=(DataNode *)node;
		  for (i=0; i<N->count; i++)
		    dest_queue->push(PQEntry((N->branch[i]).child, OBJECT, (N->branch[i].point).ToRect(), MPQ->Distance(&(N->branch[i].point), User_Distance)));
		}
	    }
	  else
	    {
	      diskacc[0]++;
	      assert(dest_top.type == OBJECT);
	      p=(dest_top.BR).ToPoint();
	      dist=dest_top.distance;
	      (res->objectList[res->numObjects]).id=(long)dest_top.n;
	      (res->objectList[res->numObjects]).point=p;
	      (res->objectList[res->numObjects]).distance=dist;
	      res->numObjects++;
	      
	      if (res->numObjects >= k)  // got all objectsa
		
		{
		  printf("with estimate:  %d dist computations instead of %d \n", num_dist_comp, max_num_dist_comp);
                  printf("size of dest queue: %d \n", dest_queue->size());
		  return res->numObjects; 
		}
	    }
	}
    }
}

// push in all entries in src_queue into dest_queue, 
// recomputing distance based on new multpoint distance function

void HTree::PushAll(PQ *src_queue, PQ *dest_queue, MultiPointQuery *mpq, float (*User_Distance) (Point *, Point *))
{
  Point object_point;
  Rect liveRegion;
  while(!(src_queue->empty()))
    {
      PQEntry top=src_queue->top();
      src_queue->pop();
      if (top.type == OBJECT)
        {
          object_point=(top.BR).ToPoint();
	  top.distance=mpq->Distance(&object_point, User_Distance);
          dest_queue->push(top);
        }
      else
        {
          if (top.els.elsvalid)
            {
              liveRegion.DecodeLiveSpace(top.els.elsarray, &(top.BR));
	      top.distance=mpq->Distance(&liveRegion, User_Distance);
              dest_queue->push(top);
            }
          else
	    {
	      top.distance=mpq->Distance(&(top.BR), User_Distance);
              dest_queue->push(top);
	    }
        }
    }
}



// push in all entries in src_queue into dest_queue,
// recomputing distance based on new multpoint distance function

void HTree::PushAll(PQ *src_queue, PQ *dest_queue, MultiPointQuery *mpq, int dist_func)
{
  Point object_point;
  Rect liveRegion;
  while(!(src_queue->empty()))
    {
      PQEntry top=src_queue->top();
      src_queue->pop();
      if (top.type == OBJECT)
        {
          object_point=(top.BR).ToPoint();
	  top.distance=mpq->Distance(&object_point, dist_func);
          dest_queue->push(top);
        }
      else
        {
          if (top.els.elsvalid)
            {
              liveRegion.DecodeLiveSpace(top.els.elsarray, &(top.BR));
	      top.distance=mpq->Distance(&liveRegion, dist_func);
              dest_queue->push(top);
            }
          else
	    {
	      top.distance=mpq->Distance(&(top.BR), dist_func);
              dest_queue->push(top);
	    }
        }
    }
}

// ------------------- FEEDBACK CODE ENDS HERE -------------------------
// ------------------- FEEDBACK CODE ENDS HERE -------------------------
// ------------------- FEEDBACK CODE ENDS HERE -------------------------


